/****************************************************************************************
*  YAROCK                                                                               *
*  Copyright (c) 2010-2013 Sebastien amardeilh <sebastien.amardeilh+yarock@gmail.com>   *
*                                                                                       *
*  This program is free software; you can redistribute it and/or modify it under        *
*  the terms of the GNU General Public License as published by the Free Software        *
*  Foundation; either version 2 of the License, or (at your option) any later           *
*  version.                                                                             *
*                                                                                       *
*  This program is distributed in the hope that it will be useful, but WITHOUT ANY      *
*  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A      *
*  PARTICULAR PURPOSE. See the GNU General Public License for more details.             *
*                                                                                       *
*  You should have received a copy of the GNU General Public License along with         *
*  this program.  If not, see <http://www.gnu.org/licenses/>.                           *
*****************************************************************************************/

#include "AlbumCover/coverprovider_lastfm.h"
#include "MediaItem/mediaitem.h"
#include "networkaccess.h"
#include "utilities.h"
#include "debug.h"

/*
********************************************************************************
*                                                                              *
*    Class CoverProvider_LastFm                                                *
*                                                                              *
********************************************************************************
*/

CoverProvider_LastFm::CoverProvider_LastFm(QObject* parent)
  : CoverProvider("last.fm", parent)
{
}

//! ----  set_search -----------------------------------------------------------
void CoverProvider_LastFm::set_search(const QString& artist, const QString& album)
{
    //Debug::debug() << " ---- CoverProvider_LastFm ->set_search  m_list_requests" << m_list_requests;

    m_album      = album;
    m_artist     = artist;
    m_cover_name = MEDIA::coverName(artist, album);
}

//! ----  start_search ---------------------------------------------------------
void CoverProvider_LastFm::start_search(int id)
{
    Debug::debug() << " ---- CoverProvider_LastFm -> start_search";
    if (m_artist.isEmpty())
    {
      emit search_finished(id);
    }
    else
    {
      slot_send_album_info_request(id);
    }
}


//! ----  slot_send_album_info_request -----------------------------------------
void CoverProvider_LastFm::slot_send_album_info_request(int id)
{
    //Debug::debug() << " ---- CoverProvider_LastFm -> slot_send_album_info_request";

    QUrl url("http://ws.audioscrobbler.com/2.0/");
    url.addQueryItem("method", "album.getinfo");
    url.addQueryItem("api_key", LASTFM_API_KEY);
    url.addQueryItem("artist", m_artist);
    url.addQueryItem("album",  m_album);

    QObject *reply = HTTP()->get(url);
    m_list_requests[reply] = id;

    connect(reply, SIGNAL(data(QByteArray)), this, SLOT(slot_parse_album_info(QByteArray)));
    connect(reply, SIGNAL(error(QNetworkReply*)), this, SLOT(slot_cover_search_stop()));
}

//! ----  slot_parse_album_info ------------------------------------------------
void CoverProvider_LastFm::slot_parse_album_info(QByteArray bytes)
{
    Debug::debug() << " ---- CoverProvider_LastFm -> slot_parse_album_info";
    /*-------------------------------------------------*/
    /* Get id from sender reply                        */
    /* ------------------------------------------------*/
    QObject* reply = qobject_cast<QObject*>(sender());

    if (!reply || !m_list_requests.contains(reply))  {
      emit search_finished(-1);
      return;
    }
    const int id = m_list_requests.take(reply);

    /*-------------------------------------------------*/
    /* Parse info                                      */
    /* ------------------------------------------------*/
    QXmlStreamReader xml(bytes);
    bool imageFound = false;

    //! Xml reading loop
    while(!xml.atEnd() && !xml.hasError())
    {
      QXmlStreamReader::TokenType token = xml.readNext();
      //! search Xml start element
      if(token == QXmlStreamReader::StartElement)
      {
        //! search Xml Image element (size = medium, large, extralarge)
        if(xml.name() == "image" && xml.attributes().value("size") == "large")
        {
          //! we have to check if image is already stored
          bool imageAlreadyPresent = false;
          const QString imageFilePath = UTIL::CONFIGDIR + "/albums/" + m_cover_name;
          if (QFile::exists(imageFilePath)) {
            Debug::debug() << " ---- CoverProvider_LastFm -> slot_parse_album_info image already exist";
            imageAlreadyPresent = true;
          }
          // image has to be stored
          if (!imageAlreadyPresent) {
            QString imageUrl = xml.readElementText();
            if (!imageUrl.isEmpty())
            {
              //Debug::debug() << " ---- CoverProvider_LastFm -> get image url";

              QUrl url = QUrl::fromEncoded(imageUrl.toUtf8());
              QObject *reply = HTTP()->get(url);
              m_list_requests[reply] = id;

              connect(reply, SIGNAL(data(QByteArray)), this, SLOT(slot_cover_received(QByteArray)));
              connect(reply, SIGNAL(error(QNetworkReply*)), this, SLOT(slot_cover_search_stop()));
              imageFound = true;
              break;
            }
          }
        } // end Xml image element
      } // end Xml start element
    } // end Xml reading loop

    // Error handling.
    if(xml.hasError()) {
      Debug::debug() << " ---- CoverProvider_LastFm -> Network error!!";
      emit search_finished(id);
    }

    if(!imageFound) {
      Debug::debug() << " ---- CoverProvider_LastFm -> No cover found!!";
      emit search_finished(id);
    }
}


//! --  slot_cover_received ----------------------------------------------------
void CoverProvider_LastFm::slot_cover_received(QByteArray bytes)
{
    //Debug::debug() << " ---- CoverProvider_LastFm -> slot_cover_received";

    // Get id from sender reply
    QObject* reply = qobject_cast<QObject*>(sender());
    if (!reply || !m_list_requests.contains(reply))   {
      emit search_finished(-1);
      return;
    }

    const int id = m_list_requests.take(reply);

    // Send signal for data reception
    emit data_received(id, bytes);
}



//! --  slot_cover_search_stop -------------------------------------------------
void CoverProvider_LastFm::slot_cover_search_stop()
{
    //Debug::debug() << " ---- CoverProvider_LastFm -> slot_cover_search_stop";

    // Get id from sender reply
    QObject* reply = qobject_cast<QObject*>(sender());
    if (!reply || !m_list_requests.contains(reply))  {
      emit search_finished(-1);
      return;
    }

    const int id = m_list_requests.take(reply);

    // Send signal for data reception
    emit search_finished(id);
}
