/****************************************************************************************
*  YAROCK                                                                               *
*  Copyright (c) 2010-2013 Sebastien amardeilh <sebastien.amardeilh+yarock@gmail.com>   *
*                                                                                       *
*  This program is free software; you can redistribute it and/or modify it under        *
*  the terms of the GNU General Public License as published by the Free Software        *
*  Foundation; either version 2 of the License, or (at your option) any later           *
*  version.                                                                             *
*                                                                                       *
*  This program is distributed in the hope that it will be useful, but WITHOUT ANY      *
*  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A      *
*  PARTICULAR PURPOSE. See the GNU General Public License for more details.             *
*                                                                                       *
*  You should have received a copy of the GNU General Public License along with         *
*  this program.  If not, see <http://www.gnu.org/licenses/>.                           *
*****************************************************************************************/


#include "networkaccess.h"
#include "constants.h"
#include "debug.h"


static const QString USER_AGENT = APP_NAME + " " + VERSION;

/*
********************************************************************************
*                                                                              *
*    Class NetworkReply                                                        *
*                                                                              *
********************************************************************************
*/
NetworkReply::NetworkReply(QNetworkReply *networkReply) : QObject(networkReply)
{
    m_networkReply = networkReply;
    m_retryCount   = 0;

    setupReply();
    
    m_timer = new QTimer(this);
    m_timer->setInterval(10000);
    m_timer->setSingleShot(true);
    connect(m_timer, SIGNAL(timeout()), SLOT(readTimeout()), Qt::UniqueConnection);
    m_timer->start();
}

void NetworkReply::setupReply() {
    connect(m_networkReply, SIGNAL(error(QNetworkReply::NetworkError)),
            SLOT(requestError(QNetworkReply::NetworkError)), Qt::UniqueConnection);
    connect(m_networkReply, SIGNAL(finished()),
            SLOT(finished()), Qt::UniqueConnection);
    connect(m_networkReply, SIGNAL(downloadProgress(qint64,qint64)),
            SLOT(downloadProgress(qint64,qint64)), Qt::UniqueConnection);
}

void NetworkReply::finished() {
    QUrl redirection = m_networkReply->attribute(QNetworkRequest::RedirectionTargetAttribute).toUrl();
    
    if (redirection.isValid()) 
    {
        if ( m_networkReply->operation() == QNetworkAccessManager::GetOperation ||
             m_networkReply->operation() == QNetworkAccessManager::HeadOperation) 
        {
            QNetworkReply *redirectReply = HTTP()->request(redirection, m_networkReply->operation());
            setParent(redirectReply);
            m_networkReply->deleteLater();
            m_networkReply = redirectReply;
            setupReply();
            m_timer->start();
            return;
        }
        else 
        {
          Debug::warning() << "Redirection not supported" << m_networkReply->url().toEncoded();
        }
    }

    emit finished(m_networkReply);
    emit data(m_networkReply->readAll());

    // bye bye my reply
    m_networkReply->deleteLater();
}

void NetworkReply::requestError(QNetworkReply::NetworkError code) 
{
    Debug::warning() << m_networkReply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt()
                     << m_networkReply->errorString() << code;
    emit error(m_networkReply);
}

void NetworkReply::downloadProgress(qint64 bytesReceived, qint64 /* bytesTotal */)
{
    // qDebug() << "Downloading" << bytesReceived << bytesTotal << m_networkReply->url();
    if (bytesReceived > 0 && m_timer->isActive()) 
    {
        m_timer->stop();
        disconnect(m_networkReply, SIGNAL(downloadProgress(qint64,qint64)),
                   this, SLOT(downloadProgress(qint64,qint64)));
    }
}

void NetworkReply::readTimeout()
{
    m_networkReply->disconnect();
    m_networkReply->abort();
    m_networkReply->deleteLater();

    if (m_networkReply->operation() != QNetworkAccessManager::GetOperation || 
        m_networkReply->operation() != QNetworkAccessManager::HeadOperation) {
        emit error(m_networkReply);
        return;
    }

    if (m_retryCount > 3) {
        emit error(m_networkReply);
        return;
    }
    
    QNetworkReply *retryReply = HTTP()->request(m_networkReply->url(), m_networkReply->operation());
    setParent(retryReply);
    m_networkReply = retryReply;
    setupReply();
    m_retryCount++;
    m_timer->start();
}


/*
********************************************************************************
*                                                                              *
*    Class NetworkAccess                                                       *
*                                                                              *
********************************************************************************
*/
NetworkAccess::NetworkAccess(QObject* parent) : QObject( parent ) {}

QNetworkRequest NetworkAccess::buildRequest(QUrl url)
{
    QNetworkRequest request(url);
    request.setRawHeader("User-Agent", USER_AGENT.toUtf8());
    request.setRawHeader("Accept-Charset", "ISO-8859-1,utf-8;q=0.7,*;q=0.7");
//     request.setRawHeader("Accept", "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8");
//     request.setRawHeader("Accept-Language", "en-us,en;q=0.5");
    request.setRawHeader("Connection", "Keep-Alive");
    return request;
}

QNetworkReply* NetworkAccess::request(QUrl url, int operation, QByteArray data) 
{
    QNetworkAccessManager *manager = NetworkManager::instance()->accessManager();

    QNetworkRequest request = buildRequest(url);

    QNetworkReply *networkReply;
    switch (operation) {

    case QNetworkAccessManager::GetOperation:
        networkReply = manager->get(request);
        break;

    case QNetworkAccessManager::HeadOperation:
        networkReply = manager->head(request);
        break;

    case QNetworkAccessManager::PostOperation:
        request.setRawHeader("Content-Type", "application/x-www-form-urlencoded");
        networkReply = manager->post(request, data);
        break;

    default:
        qWarning() << "Unknown operation:" << operation;
        return 0;
    }

    return networkReply;
}




NetworkReply* NetworkAccess::get(const QUrl url)
{
    QNetworkReply *networkReply = request(url);
    return new NetworkReply(networkReply);
}

NetworkReply* NetworkAccess::head(const QUrl url)
{
    QNetworkReply *networkReply = request(url, QNetworkAccessManager::HeadOperation);
    return new NetworkReply(networkReply);
}

NetworkReply* NetworkAccess::post(const QUrl url, QByteArray data)
{
    QNetworkReply *networkReply = request(url, QNetworkAccessManager::PostOperation,data);
    return new NetworkReply(networkReply);
}


/*
********************************************************************************
*                                                                              *
*    Class NetworkManager                                                      *
*                                                                              *
********************************************************************************
*/
NetworkManager* NetworkManager::INSTANCE = 0;

NetworkManager::NetworkManager()
{
    INSTANCE = this;
    m_network_access_manager = new QNetworkAccessManager();
}


NetworkAccess* NetworkManager::http()
{
  if (m_http.contains(QThread::currentThread()))
  {
    return m_http.value(QThread::currentThread());
  }
  else
  {
    // Debug::debug() << "NetworkAccess for thread" << QThread::currentThread();
    NetworkAccess *http = new NetworkAccess();
    m_http.insert(QThread::currentThread(), http);
    return http;
  }
}
